@@ -12,7 +12,7 @@
 #include "Python.h"
 #include "blosc.h"
 
-static int RELEASEGIL;     
+static int RELEASEGIL;
 static PyObject *BloscError;
 
 static void
@@ -57,6 +57,22 @@ PyBlosc_set_blocksize(PyObject *self, PyObject *args)
   Py_RETURN_NONE;
 }
 
+
+PyDoc_STRVAR(get_blocksize__doc__,
+"get_blocksize() -- Get the blocksize currently used.\n"
+             );
+
+static PyObject *
+PyBlosc_get_blocksize(PyObject *self)
+{
+  int blocksize;
+
+  blocksize = blosc_get_blocksize();
+
+  return Py_BuildValue("i", blocksize);
+}
+
+
 PyDoc_STRVAR(set_releasegil__doc__,
 "set_releasegil( gilstate ) -- Whether to release GIL (True) or not (False) during c-blosc calls.\n"
              );
@@ -213,25 +229,25 @@ compress_helper(void * input, size_t nbytes, size_t typesize,
   output_ptr = PyBytes_AS_STRING(output);
 
   if( RELEASEGIL )
-  { 
+  {
     // Run with GIL released, tiny overhead penalty from this (although it
-    // may be significant for smaller chunks.) 
-    
+    // may be significant for smaller chunks.)
+
     _save = PyEval_SaveThread();
     blocksize = blosc_get_blocksize();
     // if blocksize==0, blosc_compress_ctx will try to auto-optimize it.
     nthreads = blosc_get_nthreads();
     cbytes = blosc_compress_ctx(clevel, shuffle, typesize, nbytes,
-			  input, output_ptr, nbytes+BLOSC_MAX_OVERHEAD,
-               cname, blocksize, nthreads);
+                                input, output_ptr, nbytes+BLOSC_MAX_OVERHEAD,
+                                cname, blocksize, nthreads);
      PyEval_RestoreThread(_save);
      _save = NULL;
   }
   else
   { // Hold onto the Python GIL while compressing
     cbytes = blosc_compress(clevel, shuffle, typesize, nbytes,
-			  input, output_ptr,
-			  nbytes+BLOSC_MAX_OVERHEAD);
+                            input, output_ptr,
+                            nbytes+BLOSC_MAX_OVERHEAD);
   }
 
   if (cbytes < 0) {
@@ -239,8 +255,8 @@ compress_helper(void * input, size_t nbytes, size_t typesize,
     Py_DECREF(output);
     return NULL;
   }
-  
-  
+
+
   /* Attempt to resize, if it's much smaller, a copy is required. */
   if (_PyBytes_Resize(&output, cbytes) < 0){
     /* the memory exception will have been set, hopefully */
@@ -266,7 +282,7 @@ PyBlosc_compress_ptr(PyObject *self, PyObject *args)
 
   /* require an address, buffer length, typesize, clevel, shuffle and cname */
   if (!PyArg_ParseTuple(args, "Onniis:compress", &input, &nbytes,
-			&typesize, &clevel, &shuffle, &cname))
+      &typesize, &clevel, &shuffle, &cname))
     return NULL;
   /*  convert to void pointer safely */
   input_ptr = PyLong_AsVoidPtr(input);
@@ -330,6 +346,25 @@ PyBlosc_get_clib(PyObject *self, PyObject *args)
   return Py_BuildValue("s", clib);
 }
 
+PyDoc_STRVAR(get_cbuffer_sizes__doc__,
+"get_cbuffer_sizes() -- Return information about a compressed buffer,\
+the number of uncompressed and compressed bytes and the blocksize.\n"
+            );
+
+static PyObject *
+PyBlosc_get_cbuffer_sizes(PyObject *self, PyObject *args)
+{
+   void *cbuffer;
+   size_t auxbytes, nbytes, cbytes, blocksize;
+
+   if (!PyArg_ParseTuple(args, "s#:get_cbuffer_sizes", &cbuffer, &auxbytes))
+    return NULL;
+
+   blosc_cbuffer_sizes(cbuffer, &nbytes, &cbytes, &blocksize);
+
+   return Py_BuildValue("nnn", nbytes, cbytes, blocksize);
+}
+
 /*  Read blosc header from input and fetch the uncompressed size into nbytes.
  *  Also makes sure that value of the compressed bytes from the header is the
  *  same as the cbytes provided by the input.
@@ -362,13 +397,13 @@ decompress_helper(void * input, size_t nbytes, void * output)
 {
   int err, nthreads;
   PyThreadState *_save = NULL;
-  
+
   /* Do the decompression */
 //  int blosc_decompress_ctx(const void *src, void *dest, size_t destsize,
 //                         int numinternalthreads)
   if( RELEASEGIL )
-  { 
-    
+  {
+
     _save = PyEval_SaveThread();
     nthreads = blosc_get_nthreads();
     err = blosc_decompress_ctx(input, output, nbytes, nthreads);
@@ -380,16 +415,16 @@ decompress_helper(void * input, size_t nbytes, void * output)
     err = blosc_decompress(input, output, nbytes);
   }
 
-  
+
   if (err < 0) {
     blosc_error(err, "while decompressing data");
     return 0;
   }
   else if (err != (int)nbytes) {
     PyErr_Format(BloscError,
-		 "expected %d bytes of decompressed data, got %d",
-		 (int) nbytes,
-		 err);
+        "expected %d bytes of decompressed data, got %d",
+        (int) nbytes,
+        err);
     return 0;
   }
   return 1;
@@ -473,14 +508,14 @@ PyBlosc_decompress(PyObject *self, PyObject *args)
     return NULL;
   }
 
-#define branch(from_string_and_size, as_string)				 \
-  /* Book memory for the result */					 \
-    if (!(result_str = from_string_and_size(NULL, (Py_ssize_t)nbytes))){ \
-      PyBuffer_Release(&view);						 \
-      return NULL;							 \
-    }									 \
-                                                                         \
-    output = as_string(result_str);					 \
+#define branch(from_string_and_size, as_string)                               \
+  /* Book memory for the result */                                            \
+    if (!(result_str = from_string_and_size(NULL, (Py_ssize_t)nbytes))){      \
+      PyBuffer_Release(&view);                                                \
+      return NULL;                                                            \
+    }                                                                         \
+                                                                              \
+    output = as_string(result_str);                                           \
     (void)NULL
 
   if (as_bytearray) {
@@ -504,6 +539,42 @@ PyBlosc_decompress(PyObject *self, PyObject *args)
 }
 
 
+PyDoc_STRVAR(cbuffer_validate__doc__,
+"cbuffer_validate(bytesobj) -- Check if compressed data is safe.\n\n"
+"Checks that it is safe to attempt decompression of compressed data.\n"
+"This does not guarantee that decompression will be successful,\n"
+"only that it is safe to attempt decompression.");
+
+static PyObject *
+PyBlosc_cbuffer_validate(PyObject *self, PyObject *args)
+{
+  Py_buffer view;
+  void *input;
+  size_t nbytes, cbytes;
+  int result;
+
+  #if PY_MAJOR_VERSION <= 2
+    if (!PyArg_ParseTuple(args, "s*:cbuffer_validate", &view))
+      return NULL;
+
+  #elif PY_MAJOR_VERSION >= 3
+    /* y* :bytes like object EXCLUDING unicode and anything that supports
+     * the buffer interface. This is the recommended way to accept binary
+     * data in Python 3. */
+    if (!PyArg_ParseTuple(args, "y*:cbuffer_validate", &view))
+      return NULL;
+  #endif
+  cbytes = view.len;
+  input = view.buf;
+  result = blosc_cbuffer_validate(input, cbytes, &nbytes);
+  if (result == 0) {
+    Py_RETURN_TRUE;
+  }else{
+    Py_RETURN_FALSE;
+  }
+}
+
+
 static PyMethodDef blosc_methods[] =
 {
   {"compress", (PyCFunction)PyBlosc_compress,  METH_VARARGS,
@@ -520,6 +591,8 @@ static PyMethodDef blosc_methods[] =
    set_nthreads__doc__},
   {"set_blocksize", (PyCFunction)PyBlosc_set_blocksize, METH_VARARGS,
    set_blocksize__doc__},
+  {"get_blocksize", (PyCFunction)PyBlosc_get_blocksize, METH_VARARGS,
+   get_blocksize__doc__},
   {"set_releasegil", (PyCFunction)PyBlosc_set_releasegil, METH_VARARGS,
    set_releasegil__doc__},
   {"compressor_list", (PyCFunction)PyBlosc_compressor_list, METH_VARARGS,
@@ -532,6 +605,10 @@ static PyMethodDef blosc_methods[] =
    clib_info__doc__},
   {"get_clib", (PyCFunction)PyBlosc_get_clib, METH_VARARGS,
    get_clib__doc__},
+  {"get_cbuffer_sizes", (PyCFunction)PyBlosc_get_cbuffer_sizes, METH_VARARGS,
+   get_cbuffer_sizes__doc__},
+  {"cbuffer_validate", (PyCFunction)PyBlosc_cbuffer_validate, METH_VARARGS,
+   cbuffer_validate__doc__},
   {"init", (PyCFunction)PyBlosc_init, METH_VARARGS,
    init__doc__},
   {"destroy", (PyCFunction)PyBlosc_destroy, METH_VARARGS,
@@ -587,6 +664,13 @@ PyMODINIT_FUNC
 PyInit_blosc_extension(void) {
   PyObject *m = PyModule_Create(&blosc_def);
 
+
+  BloscError = PyErr_NewException("blosc_extension.error", NULL, NULL);
+  if (BloscError != NULL) {
+    Py_INCREF(BloscError);
+    PyModule_AddObject(m, "error", BloscError);
+  }
+
   /* Integer macros */
   PyModule_AddIntMacro(m, BLOSC_MAX_BUFFERSIZE);
   PyModule_AddIntMacro(m, BLOSC_MAX_THREADS);
@@ -602,3 +686,5 @@ PyInit_blosc_extension(void) {
   return m;
 }
 #endif
+
+
